#!/usr/bin/env python

"""
Webware classes.
"""

import Generic
from cgi import parse_qs
import StringIO
from Helpers import Environment
from Helpers.Request import Cookie
from Helpers.Response import ConvertingStream

class Transaction(Generic.Transaction):

    """
    Webware transaction interface.
    """

    def __init__(self, trans):

        "Initialise the transaction using the Webware transaction 'trans'."

        self.trans = trans
        self.user = None
        self.content_type = None

    # Request-related methods.

    def get_request_stream(self):

        """
        Returns the request stream for the transaction.
        """

        request = self.trans.request()
        try:
            stream = request.rawInput(rewind=1)
            if stream is None:
                return StringIO.StringIO("")

        # NOTE: Dubious catch-all situation, but it is difficult to control
        # NOTE: cases where Webware's internals themselves fail.

        except:
            return StringIO.StringIO("")

        return stream

    def get_request_method(self):

        """
        Returns the request method.
        """

        return self.trans.request().method()

    def get_headers(self):

        """
        Returns all request headers as a dictionary-like object mapping header
        names to values.

        NOTE: If duplicate header names are permitted, then this interface will
        NOTE: need to change.
        """

        # Use the Webware environment and some assumptions about variable names.
        # NOTE: Using lower case for the header names.

        env = self.trans.request().environ()
        return Environment.get_headers(env)

    def get_header_values(self, key):

        """
        Returns a list of all request header values associated with the given
        'key'. Note that according to RFC 2616, 'key' is treated as a
        case-insensitive string.
        """

        # Use the Webware environment and some assumptions about variable names.

        env = self.trans.request().environ()
        cgi_key = "HTTP_" + key.replace("-", "_").upper()
        if env.has_key(cgi_key):
            return [env[cgi_key]]
        else:
            return []

    def get_content_type(self):

        """
        Returns the content type specified on the request, along with the
        charset employed.
        """

        return self.parse_content_type(self.trans.request().contentType())

    def get_content_charsets(self):

        """
        Returns the character set preferences.
        NOTE: Requires enhancements to HTTPRequest.
        """

        return self.trans.request().contentCharsets()

    def get_content_languages(self):

        """
        Returns extracted language information from the transaction.
        NOTE: Requires enhancements to HTTPRequest.
        """

        return self.trans.request().contentLanguages()

    def get_path(self):

        """
        Returns the entire path from the request.
        """

        return self.trans.request().uri()

    def get_path_without_query(self):

        """
        Returns the entire path from the request minus the query string.
        """

        return self.get_path().split("?")[0]

    def get_path_info(self):

        """
        Returns the "path info" (the part of the URL after the resource name
        handling the current request) from the request.
        """

        return self.trans.request().extraURLPath()

    def get_query_string(self):

        """
        Returns the query string from the path in the request.
        """

        return self.trans.request().queryString()

    # Higher level request-related methods.

    def get_fields_from_path(self):

        """
        Extracts the form fields from the path specified in the transaction. The
        underlying framework may refuse to supply fields from the path if
        handling a POST transaction.

        Returns a dictionary mapping field names to lists of values (even if a
        single value is associated with any given field name).
        """

        return parse_qs(self.get_query_string(), keep_blank_values=1)

    def get_fields_from_body(self, encoding=None):

        """
        Extracts the form fields from the message body in the transaction. The
        optional 'encoding' parameter specifies the character encoding of the
        message body for cases where no such information is available, but where
        the default encoding is to be overridden.

        Returns a dictionary mapping field names to lists of values (even if a
        single value is associated with any given field name). Each value is
        either a Unicode object (representing a simple form field, for example)
        or a file-like object (representing a file upload form field, for
        example).
        """

        encoding = self.get_content_type().charset or encoding or "iso-8859-1"

        # Fix the non-list results.

        fields = {}
        for field_name, field_values in self.trans.request().fields().items():
            if type(field_values) == type([]):
                fields[field_name] = []
                for field_value in field_values:
                    if hasattr(field_value, "file"):
                        fields[field_name].append(field_value.file)
                    else:
                        fields[field_name].append(unicode(field_value, encoding))
            else:
                if hasattr(field_values, "file"):
                    fields[field_name] = [field_values.file]
                else:
                    fields[field_name] = [unicode(field_values, encoding)]
        return fields

    def get_user(self):

        """
        Extracts user information from the transaction.

        Returns a username as a string or None if no user is defined.
        """

        # NOTE: Webware relies entirely on a CGI-style environment where the
        # NOTE: actual headers are not available. Therefore, the Web server must
        # NOTE: itself be set up to provide user support.

        if self.user is not None:
            return self.user

        try:
            return self.trans.request().remoteUser()
        except KeyError, exc:
            return None

    def get_cookies(self):

        """
        Obtains cookie information from the request.

        Returns a dictionary mapping cookie names to cookie objects.
        """

        cookies = {}
        for name, value in self.trans.request().cookies().items():
            cookies[name] = Cookie(name, value)
        return cookies

    def get_cookie(self, cookie_name):

        """
        Obtains cookie information from the request.

        Returns a cookie object for the given 'cookie_name' or None if no such
        cookie exists.
        """

        try:
            return Cookie(cookie_name, self.trans.request().cookie(cookie_name))
        except KeyError:
            return None

    # Response-related methods.

    def get_response_stream(self):

        """
        Returns the response stream for the transaction.
        """

        # Unicode can upset this operation. Using either the specified charset,
        # the same charset as that used in the request, or a default encoding.

        encoding = self.get_content_type().charset or "utf-8"
        if self.content_type:
            encoding = self.content_type.charset or encoding
        return ConvertingStream(self.trans.response(), encoding)

    def get_response_code(self):

        """
        Get the response code associated with the transaction. If no response
        code is defined, None is returned.
        """

        # NOTE: Webware treats the response code as just another header.

        status = self.trans.response().header("Status", None)
        try:
            if status is not None:
                return int(status)
            else:
                return None
        except ValueError:
            return None

    def set_response_code(self, response_code):

        """
        Set the 'response_code' using a numeric constant defined in the HTTP
        specification.
        """

        self.trans.response().setStatus(response_code)

    def set_header_value(self, header, value):

        """
        Set the HTTP 'header' with the given 'value'.
        """

        self.trans.response().setHeader(self.format_header_value(header), self.format_header_value(value))

    def set_content_type(self, content_type):

        """
        Sets the 'content_type' for the response.
        """

        # Remember the content type for encoding purposes later.

        self.content_type = content_type
        return self.trans.response().setHeader("Content-Type", self.format_content_type(content_type))

    # Higher level response-related methods.

    def set_cookie(self, cookie):

        """
        Stores the given 'cookie' object in the response.
        """

        self.trans.response().addCookie(cookie)

    def set_cookie_value(self, name, value, path=None, expires=None):

        """
        Stores a cookie with the given 'name' and 'value' in the response.

        The optional 'path' is a string which specifies the scope of the cookie,
        and the optional 'expires' parameter is a value compatible with the
        time.time function, and indicates the expiry date/time of the cookie.
        """

        self.trans.response().setCookie(name, value, path, expires)

    def delete_cookie(self, cookie_name):

        """
        Adds to the response a request that the cookie with the given
        'cookie_name' be deleted/discarded by the client.
        """

        self.trans.response().delCookie(cookie_name)

    # Session-related methods.

    def get_session(self, create=1):

        """
        Gets a session corresponding to an identifier supplied in the
        transaction.

        If no session has yet been established according to information
        provided in the transaction then the optional 'create' parameter
        determines whether a new session will be established.

        Where no session has been established and where 'create' is set to 0
        then None is returned. In all other cases, a session object is created
        (where appropriate) and returned.
        """

        # NOTE: Should really use Webware's hasSession method.

        session = self.trans.session()
        return Session(session)

    def expire_session(self):

        """
        Expires any session established according to information provided in the
        transaction.
        """

        self.trans.request().setSessionExpired(1)

    # Application-specific methods.

    def set_user(self, username):

        """
        An application-specific method which sets the user information with
        'username' in the transaction. This affects subsequent calls to
        'get_user'.
        """

        self.user = username

class Session:

    "A more dictionary-like session object than the one Webware provides."

    def __init__(self, session):
        self.session = session

    def items(self):
        return self.session.values().items()

    def __getattr__(self, name):
        return getattr(self.__dict__["session"], name)

# vim: tabstop=4 expandtab shiftwidth=4
